"""BinaryOutput Set Output request message."""

from typing import List, Optional, Union
from iqrfpy.enums.commands import BinaryOutputRequestCommands
from iqrfpy.enums.message_types import BinaryOutputMessages
from iqrfpy.enums.peripherals import Standards
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.binaryoutput.requests.binary_output_state import BinaryOutputState
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['SetOutputRequest']


class SetOutputRequest(IRequest):
    """BinaryOutput Set Output request class."""

    __slots__ = ('_binouts',)

    def __init__(self, nadr: int, binouts: Optional[List[BinaryOutputState]] = None,
                 hwpid: int = dpa_constants.HWPID_MAX, dpa_rsp_time: Optional[float] = None,
                 dev_process_time: Optional[float] = None, msgid: Optional[str] = None):
        """Enumerate request constructor.

        Args:
            nadr (int): Network address
            binouts (List[BinaryOutputState], optional): Binary output states to set.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        self._validate_binouts(binouts=binouts)
        super().__init__(
            nadr=nadr,
            pnum=Standards.BINARY_OUTPUT,
            pcmd=BinaryOutputRequestCommands.SET_OUTPUT,
            m_type=BinaryOutputMessages.SET_OUTPUT,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )
        self._binouts: List[BinaryOutputState] = binouts if binouts is not None else []
        self._binouts.sort(key=lambda x: x.index)

    @staticmethod
    def _validate_binouts(binouts: Optional[List[BinaryOutputState]] = None):
        """Validate binary output states parameter.

        Args:
            binouts (List[BinaryOutputState], optional): Binary output states to set.
        Raises:
            RequestParameterInvalidValueError: If binouts contains more than 32 values.
        """
        if binouts is None or len(binouts) == 0:
            return
        if len(binouts) > 32:
            raise RequestParameterInvalidValueError('Binouts length should be at most 32 values.')

    @property
    def binouts(self) -> Optional[List[BinaryOutputState]]:
        """:obj:`list` of :obj:`BinaryOutputState` or :obj:`None`: Binary output states to set.

        Getter and setter.
        """
        return self._binouts

    @binouts.setter
    def binouts(self, value: Optional[List[BinaryOutputState]] = None):
        self._validate_binouts(value)
        self._binouts = value if value is not None else []
        self._binouts.sort(key=lambda x: x.index)

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        binouts = sorted(self._binouts, key=lambda x: x.index)
        indexes = [bo.index for bo in binouts]
        states = [bo.state for bo in binouts]
        self._pdata = Common.indexes_to_4byte_bitmap(indexes) + states
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        self._params = {
            'binOuts': [bo.to_json() for bo in self.binouts]
        }
        return super().to_json()



"""IO Get request message."""

from typing import Optional, Union
from iqrfpy.enums.commands import IORequestCommands
from iqrfpy.enums.message_types import IOMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['GetRequest']


class GetRequest(IRequest):
    """IO Get request class."""

    def __init__(self, nadr: int, hwpid: int = dpa_constants.HWPID_MAX,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Get request constructor.

        Args:
            nadr (int): Device address.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.IO,
            pcmd=IORequestCommands.GET,
            m_type=IOMessages.GET,
            hwpid=hwpid,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        return super().to_json()



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import CoordinatorResponseCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.coordinator.responses.backup import BackupResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': CoordinatorMessages.BACKUP,
    'msgid': 'backupTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 90,
    'result': {
        'networkData': [104, 194, 100, 134, 122, 60, 24, 180, 62, 243, 129, 4, 150, 214, 237, 148, 39, 37, 89, 220, 245,
                        188, 20, 151, 251, 68, 134, 102, 163, 173, 4, 167, 252, 14, 229, 152, 137, 194, 182, 75, 35,
                        146, 93, 222, 170, 169, 163, 197, 168]
    },
    'dpa': b'\x00\x00\x00\x8b\x00\x00\x00\x5a\x68\xc2\x64\x86\x7a\x3c\x18\xb4\x3e\xf3\x81\x04\x96\xd6\xed\x94\x27\x25'
           b'\x59\xdc\xf5\xbc\x14\x97\xfb\x44\x86\x66\xa3\xad\x04\xa7\xfc\x0e\xe5\x98\x89\xc2\xb6\x4b\x23\x92\x5d\xde'
           b'\xaa\xa9\xa3\xc5\xa8'
}

data_ok_1: dict = {
    'mtype': CoordinatorMessages.BACKUP,
    'msgid': 'backupTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'networkData': [76, 31, 9, 191, 70, 60, 78, 16, 207, 238, 130, 216, 94, 140, 209, 39, 180, 131, 226, 181, 23, 1,
                        21, 176, 209, 24, 61, 237, 237, 106, 169, 130, 135, 53, 161, 78, 237, 227, 223, 128, 48, 102,
                        204, 121, 147, 65, 191, 215, 98],
    },
    'dpa': b'\x00\x00\x00\x8b\x02\x04\x00\x23\x4c\x1f\x09\xbf\x46\x3c\x4e\x10\xcf\xee\x82\xd8\x5e\x8c\xd1\x27\xb4\x83'
           b'\xe2\xb5\x17\x01\x15\xb0\xd1\x18\x3d\xed\xed\x6a\xa9\x82\x87\x35\xa1\x4e\xed\xe3\xdf\x80\x30\x66\xcc\x79'
           b'\x93\x41\xbf\xd7\x62'
}

data_error: dict = {
    'mtype': CoordinatorMessages.BACKUP,
    'msgid': 'backupTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x8b\x04\x04\x01\x23'
}


class BackupResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, BackupResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, BackupResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, BackupResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, BackupResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, BackupResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, BackupResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.COORDINATOR)
        with self.subTest():
            self.assertEqual(response.pcmd, CoordinatorResponseCommands.BACKUP)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, CoordinatorMessages.BACKUP)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            BackupResponse.from_dpa(b'\x00\x00\x00\x89\x00\x00\x00\x40')

    @parameterized.expand([
        ['from_dpa', data_ok['result']['networkData'], BackupResponse.from_dpa(data_ok['dpa'])],
        ['from_dpa', data_ok_1['result']['networkData'], BackupResponse.from_dpa(data_ok_1['dpa'])],
        ['from_json', data_ok['result']['networkData'], BackupResponse.from_json(generate_json_response(data_ok))],
        ['from_json', data_ok_1['result']['networkData'], BackupResponse.from_json(generate_json_response(data_ok_1))]
    ])
    def test_get_network_data(self, _, network_data, response: BackupResponse):
        self.assertEqual(response.network_data, network_data)

    @parameterized.expand([
        ['from_dpa_error', BackupResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', BackupResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_network_data_error(self, _, response: BackupResponse):
        self.assertIsNone(response.network_data)



"""Quantity data module.

This module contains Quantity dataclasses used when parsing Sensor data.
"""

from dataclasses import dataclass
from typing import Union
from iqrfpy.exceptions import UnknownSensorTypeError
from iqrfpy.utils.sensor_constants import SensorTypes, SensorFrcCommands

__all__ = [
    'Temperature',
    'CarbonDioxide',
    'VolatileOrganicCompound',
    'ExtraLowVoltage',
    'EarthsMagneticField',
    'LowVoltage',
    'Current',
    'Power',
    'MainsFrequency',
    'TimeSpan',
    'Illuminance',
    'NitrogenDioxide',
    'SulfurDioxide',
    'CarbonMonoxide',
    'Ozone',
    'AtmosphericPressure',
    'ColorTemperature',
    'ParticulatesPM25',
    'SoundPressureLevel',
    'Altitude',
    'Acceleration',
    'Ammonia',
    'Methane',
    'ShortLength',
    'ParticulatesPM1',
    'ParticulatesPM4',
    'ParticulatesPM10',
    'TotalVolatileOrganicCompound',
    'NitrogenOxides',
    'RelativeHumidity',
    'BinaryData7',
    'PowerFactor',
    'UVIndex',
    'PH',
    'RSSI',
    'BinaryData30',
    'Consumption',
    'Datetime',
    'TimeSpanLong',
    'Latitude',
    'Longitude',
    'DataBlock',
    'get_sensor_class'
]


@dataclass
class Temperature:
    """Temperature dataclass."""

    type = SensorTypes.TEMPERATURE
    name = 'Temperature'
    short_name = 'T'
    unit = '˚C'
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE,
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class CarbonDioxide:
    """Carbon dioxide dataclass."""

    type = SensorTypes.CARBON_DIOXIDE
    name = 'Carbon dioxide'
    short_name = 'CO2'
    unit = 'ppm'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE,
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class VolatileOrganicCompound:
    """Volatile organic compound dataclass."""

    type = SensorTypes.VOLATILE_ORGANIC_COMPOUND
    name = 'Volatile organic compound'
    short_name = 'VOC'
    unit = 'ppm'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE,
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ExtraLowVoltage:
    """Extra-low voltage dataclass."""

    type = SensorTypes.EXTRA_LOW_VOLTAGE
    name = 'Extra-low voltage'
    short_name = 'U'
    unit = 'V'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class EarthsMagneticField:
    """Earth's magnetic field dataclass."""

    type = SensorTypes.EARTHS_MAGNETIC_FIELD
    name = "Earth's magnetic field"
    short_name = 'B'
    unit = 'T'
    decimal_places = 7
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class LowVoltage:
    """Low voltage dataclass."""

    type = SensorTypes.LOW_VOLTAGE
    name = 'Low voltage'
    short_name = 'U'
    unit = 'V'
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Current:
    """Current dataclass."""

    type = SensorTypes.CURRENT
    name = 'Current'
    short_name = 'I'
    unit = 'A'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Power:
    """Power dataclass."""

    type = SensorTypes.POWER
    name = 'Power'
    short_name = 'P'
    unit = 'W'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class MainsFrequency:
    """Mains frequency dataclass."""

    type = SensorTypes.MAINS_FREQUENCY
    name = 'Mains frequency'
    short_name = 'f'
    unit = 'Hz'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class TimeSpan:
    """Timespan dataclass."""

    type = SensorTypes.TIME_SPAN
    name = 'Timespan'
    short_name = 't'
    unit = 's'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Illuminance:
    """Illuminance dataclass."""

    type = SensorTypes.ILLUMINANCE
    name = 'Illuminance'
    short_name = 'Ev'
    unit = 'lx'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class NitrogenDioxide:
    """Nitrogen dioxide dataclass."""

    type = SensorTypes.NITROGEN_DIOXIDE
    name = 'Nitrogen dioxide'
    short_name = 'NO2'
    unit = 'ppm'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class SulfurDioxide:
    """Sulfur dioxide dataclass."""

    type = SensorTypes.SULFUR_DIOXIDE
    name = 'Sulfur dioxide'
    short_name = 'SO2'
    unit = 'ppm'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class CarbonMonoxide:
    """Carbon monoxide dataclass."""

    type = SensorTypes.CARBON_MONOXIDE
    name = 'Carbon monoxide'
    short_name = 'CO'
    unit = 'ppm'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Ozone:
    """Ozone dataclass."""

    type = SensorTypes.OZONE
    name = 'Ozone'
    short_name = 'O3'
    unit = 'ppm'
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class AtmosphericPressure:
    """Atmospheric pressure dataclass."""

    type = SensorTypes.ATMOSPHERIC_PRESSURE
    name = 'Atmospheric pressure'
    short_name = 'p'
    unit = 'hPa'
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ColorTemperature:
    """Color temperature dataclass."""

    type = SensorTypes.COLOR_TEMPERATURE
    name = 'Color temperature'
    short_name = 'Tc'
    unit = 'K'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ParticulatesPM25:
    """Particulates PM2.5 dataclass."""

    type = SensorTypes.PARTICULATES_PM2_5
    name = 'Particulates PM2.5'
    short_name = 'PM2.5'
    unit = 'µg/m3'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class SoundPressureLevel:
    """Sound pressure level dataclass."""

    type = SensorTypes.SOUND_PRESSURE_LEVEL
    name = 'Sound pressure level'
    short_name = 'Lp'
    unit = 'dB'
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Altitude:
    """Altitude dataclass."""

    type = SensorTypes.ALTITUDE
    name = 'Altitude'
    short_name = 'h'
    unit = 'm'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Acceleration:
    """Acceleration dataclass."""

    types = SensorTypes.ACCELERATION
    name = 'Acceleration'
    short_name = 'a'
    unit = 'm/s2'
    decimal_places = 8
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Ammonia:
    """Ammonia dataclass."""

    types = SensorTypes.AMMONIA
    name = 'Ammonia'
    short_name = 'NH3'
    unit = 'ppm'
    decimal_places = 1
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class Methane:
    """Methane dataclass."""

    type = SensorTypes.METHANE
    name = 'Methane'
    short_name = 'CH4'
    unit = '%'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ShortLength:
    """Short length dataclass."""

    type = SensorTypes.SHORT_LENGTH
    name = 'Short length'
    short_name = 'l'
    unit = 'm'
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ParticulatesPM1:
    """Particulates PM1 dataclass."""

    type = SensorTypes.PARTICULATES_PM1
    name = 'Particulates PM1'
    short_name = 'PM1'
    unit = 'µg/m3'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ParticulatesPM4:
    """Particulates PM4 dataclass."""

    type = SensorTypes.PARTICULATES_PM4
    name = 'Particulates PM4'
    short_name = 'PM4'
    unit = 'µg/m3'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class ParticulatesPM10:
    """Particulates PM10 dataclass."""

    type = SensorTypes.PARTICULATES_PM10
    name = 'Particulates PM10'
    short_name = 'PM10'
    unit = 'µg/m3'
    decimal_places = 2
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class TotalVolatileOrganicCompound:
    """Total volatile organic compound dataclass."""

    type = SensorTypes.TOTAL_VOLATILE_ORGANIC_COMPOUND
    name = 'Total volatile organic compound'
    short_name = 'TVOC'
    unit = 'µg/m3'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class NitrogenOxides:
    """Nitrogen oxides dataclass."""

    type = SensorTypes.NITROGEN_OXIDES
    name = 'Nitrogen oxides'
    short_name = 'NOX'
    unit = ''
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES
    ]


@dataclass
class RelativeHumidity:
    """Relative humidity dataclass."""

    type = SensorTypes.RELATIVE_HUMIDITY
    name = 'Relative humidity'
    short_name = 'RH'
    unit = '%'
    decimal_places = 1
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE
    ]


@dataclass
class BinaryData7:
    """Binary data 7 dataclass."""

    type = SensorTypes.BINARY_DATA7
    name = 'Binary data7'
    short_name = 'bin7'
    unit = ''
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BITS,
        SensorFrcCommands.FRC_1BYTE
    ]


@dataclass
class PowerFactor:
    """Power factor dataclass."""

    type = SensorTypes.POWER_FACTOR
    name = 'Power factor'
    short_name = 'cos θ'
    unit = ''
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE
    ]


@dataclass
class UVIndex:
    """UV index dataclass."""

    type = SensorTypes.UV_INDEX
    name = 'UV index'
    short_name = 'UV'
    unit = ''
    decimal_places = 3
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE
    ]


@dataclass
class PH:
    """PH dataclass."""

    type = SensorTypes.PH
    name = 'pH'
    short_name = 'pH'
    unit = ''
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE
    ]


@dataclass
class RSSI:
    """RSSI dataclass."""

    type = SensorTypes.RSSI
    name = 'RSSI'
    short_name = 'RSSI'
    unit = 'dBm'
    decimal_places = 1
    frc_commands = [
        SensorFrcCommands.FRC_1BYTE
    ]


@dataclass
class BinaryData30:
    """Binary data 30 dataclass."""

    type = SensorTypes.BINARY_DATA30
    name = 'Binary data30'
    short_name = 'bin30'
    unit = ''
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_2BYTES,
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class Consumption:
    """Consumption dataclass."""

    type = SensorTypes.CONSUMPTION
    name = 'Consumption'
    short_name = 'E'
    unit = 'Wh'
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class Datetime:
    """Datetime dataclass."""

    type = SensorTypes.DATETIME
    name = 'DateTime'
    short_name = 'DateTime'
    unit = ''
    decimal_places = 0
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class TimeSpanLong:
    """Timespan long dataclass."""

    type = SensorTypes.TIME_SPAN_LONG
    name = 'Timespan long'
    short_name = 't'
    unit = 's'
    decimal_places = 4
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class Latitude:
    """Latitude dataclass."""

    type = SensorTypes.LATITUDE
    name = 'Latitude'
    short_name = 'LAT'
    unit = '°'
    decimal_places = 7
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class Longitude:
    """Longitude dataclass."""

    type = SensorTypes.LONGITUDE
    name = 'Longitude'
    short_name = 'LONG'
    unit = '°'
    decimal_places = 7
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class TemperatureFloat:
    """Temperature float dataclass."""

    type = SensorTypes.TEMPERATURE_FLOAT
    name = 'Temperature'
    short_name = 'T'
    unit = '˚C'
    decimal_places = 7
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class Length:
    """Length dataclass."""

    type = SensorTypes.LENGTH
    name = 'Length'
    short_name = 'l'
    unit = 'm'
    decimal_places = 7
    frc_commands = [
        SensorFrcCommands.FRC_4BYTES
    ]


@dataclass
class DataBlock:
    """Data block dataclass."""

    type = SensorTypes.DATA_BLOCK
    name = 'Data block'
    short_name = 'datablock'
    unit = ''
    decimal_places = 0
    frc_commands = []


_type_classes = {
    SensorTypes.TEMPERATURE: Temperature,
    SensorTypes.CARBON_DIOXIDE: CarbonDioxide,
    SensorTypes.VOLATILE_ORGANIC_COMPOUND: VolatileOrganicCompound,
    SensorTypes.EXTRA_LOW_VOLTAGE: ExtraLowVoltage,
    SensorTypes.EARTHS_MAGNETIC_FIELD: EarthsMagneticField,
    SensorTypes.LOW_VOLTAGE: LowVoltage,
    SensorTypes.CURRENT: Current,
    SensorTypes.POWER: Power,
    SensorTypes.MAINS_FREQUENCY: MainsFrequency,
    SensorTypes.TIME_SPAN: TimeSpan,
    SensorTypes.ILLUMINANCE: Illuminance,
    SensorTypes.NITROGEN_DIOXIDE: NitrogenDioxide,
    SensorTypes.SULFUR_DIOXIDE: SulfurDioxide,
    SensorTypes.CARBON_MONOXIDE: CarbonMonoxide,
    SensorTypes.OZONE: Ozone,
    SensorTypes.ATMOSPHERIC_PRESSURE: AtmosphericPressure,
    SensorTypes.COLOR_TEMPERATURE: ColorTemperature,
    SensorTypes.PARTICULATES_PM2_5: ParticulatesPM25,
    SensorTypes.SOUND_PRESSURE_LEVEL: SoundPressureLevel,
    SensorTypes.ALTITUDE: Altitude,
    SensorTypes.ACCELERATION: Acceleration,
    SensorTypes.AMMONIA: Ammonia,
    SensorTypes.METHANE: Methane,
    SensorTypes.SHORT_LENGTH: ShortLength,
    SensorTypes.PARTICULATES_PM1: ParticulatesPM1,
    SensorTypes.PARTICULATES_PM4: ParticulatesPM4,
    SensorTypes.PARTICULATES_PM10: ParticulatesPM10,
    SensorTypes.TOTAL_VOLATILE_ORGANIC_COMPOUND: TotalVolatileOrganicCompound,
    SensorTypes.NITROGEN_OXIDES: NitrogenOxides,
    SensorTypes.RELATIVE_HUMIDITY: RelativeHumidity,
    SensorTypes.BINARY_DATA7: BinaryData7,
    SensorTypes.POWER_FACTOR: PowerFactor,
    SensorTypes.UV_INDEX: UVIndex,
    SensorTypes.PH: PH,
    SensorTypes.RSSI: RSSI,
    SensorTypes.BINARY_DATA30: BinaryData30,
    SensorTypes.CONSUMPTION: Consumption,
    SensorTypes.DATETIME: Datetime,
    SensorTypes.TIME_SPAN_LONG: TimeSpanLong,
    SensorTypes.LATITUDE: Latitude,
    SensorTypes.LONGITUDE: Longitude,
    SensorTypes.TEMPERATURE_FLOAT: TemperatureFloat,
    SensorTypes.LENGTH: Length,
    SensorTypes.DATA_BLOCK: DataBlock,
}


def get_sensor_class(sensor_type: Union[SensorTypes, int]):
    """Return quantity class corresponding to sensor type.

    Args:
        sensor_type (Union[SensorTypes, int]): Sensor type (represents a quantity)
    Returns:
        Quantity dataclass
    Raises:
        UnknownSensorTypeError: Raised if sensor type is passed as integer and the value is not recognized
        ValueError: Raised if sensor type is recognized, but corresponding quantity data is missing
    """
    if not SensorTypes.has_value(sensor_type):
        raise UnknownSensorTypeError(f'Unknown or unsupported sensor type: {sensor_type}')
    if sensor_type not in _type_classes:
        raise ValueError(f'Quantity data not available for sensor type: {sensor_type}')
    return _type_classes[sensor_type]



import unittest
from typing import List, Union
from parameterized import parameterized
from iqrfpy.utils.quantity_data import Temperature, BinaryData7, RelativeHumidity, DataBlock
from iqrfpy.utils.sensor_parser import SensorParser, SensorData
from iqrfpy.utils.sensor_constants import SensorTypes, SensorFrcCommands, SensorFrcErrors


class SensorParserTestCase(unittest.TestCase):

    @parameterized.expand([
        [
            [SensorTypes.TEMPERATURE, SensorTypes.BINARY_DATA7, SensorTypes.BINARY_DATA7],
            [0, 128, 35, 127],
            [
                SensorData(
                    sensor_type=SensorTypes.TEMPERATURE,
                    index=0,
                    name=Temperature.name,
                    short_name=Temperature.short_name,
                    unit=Temperature.unit,
                    decimal_places=Temperature.decimal_places,
                    frc_commands=Temperature.frc_commands,
                    value=None
                ),
                SensorData(
                    sensor_type=SensorTypes.BINARY_DATA7,
                    index=1,
                    name=BinaryData7.name,
                    short_name=BinaryData7.short_name,
                    unit=BinaryData7.unit,
                    decimal_places=BinaryData7.decimal_places,
                    frc_commands=BinaryData7.frc_commands,
                    value=35
                ),
                SensorData(
                    sensor_type=SensorTypes.BINARY_DATA7,
                    index=2,
                    name=BinaryData7.name,
                    short_name=BinaryData7.short_name,
                    unit=BinaryData7.unit,
                    decimal_places=BinaryData7.decimal_places,
                    frc_commands=BinaryData7.frc_commands,
                    value=127
                )
            ]
        ],
        [
            [SensorTypes.RELATIVE_HUMIDITY, SensorTypes.DATA_BLOCK],
            [10, 2, 4, 4],
            [
                SensorData(
                    sensor_type=SensorTypes.RELATIVE_HUMIDITY,
                    index=0,
                    name=RelativeHumidity.name,
                    short_name=RelativeHumidity.short_name,
                    unit=RelativeHumidity.unit,
                    decimal_places=RelativeHumidity.decimal_places,
                    frc_commands=RelativeHumidity.frc_commands,
                    value=5
                ),
                SensorData(
                    sensor_type=SensorTypes.DATA_BLOCK,
                    index=1,
                    name=DataBlock.name,
                    short_name=DataBlock.short_name,
                    unit=DataBlock.unit,
                    decimal_places=DataBlock.decimal_places,
                    frc_commands=DataBlock.frc_commands,
                    value=[4, 4]
                )
            ]
        ]
    ])
    def test_read_sensors_dpa(self, sensor_types: List[SensorTypes], data: List[int], results: List[SensorData]):
        sensor_data = SensorParser.read_sensors_dpa(sensor_types=sensor_types, dpa=data)
        for i in range(len(results)):
            self.assertEqual(sensor_data[i].value, results[i].value)

    @parameterized.expand([
        [
            'unknown_type',
            [30],
            [10]
        ],
        [
            'too_little_sensor_types_for_data',
            [1],
            [10, 20, 30]
        ],
        [
            'too_many_sensor_types_for_data',
            [1, 1, 1],
            [10, 20, 30]
        ],
        [
            'data_block_invalid_data_len',
            [1, 192],
            [10, 20, 5, 10, 20]
        ],
        [
            'data_block_invalid_data_len_2',
            [192],
            [3, 12, 17]
        ]
    ])
    def test_read_sensors_dpa_invalid(self, _, sensor_types: List[SensorTypes], data: List[int]):
        with self.assertRaises(ValueError):
            SensorParser.read_sensors_dpa(sensor_types=sensor_types, dpa=data)

    @parameterized.expand([
        [
            '2bit_frc_bin7_no_count',
            SensorTypes.BINARY_DATA7,
            33,
            SensorFrcCommands.FRC_2BITS,
            [14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            None,
            [1, SensorFrcErrors.FRC_NOT_IMPLEMENTED, 1] + [SensorFrcErrors.NO_FRC_RESPONSE] * 236
        ],
        [
            '2bit_bin7_count',
            SensorTypes.BINARY_DATA7,
            33,
            SensorFrcCommands.FRC_2BITS,
            [14, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 10, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            3,
            [1, SensorFrcErrors.FRC_NOT_IMPLEMENTED, 1]
        ],
        [
            '1byte_temperature_no_count',
            SensorTypes.TEMPERATURE,
            0,
            SensorFrcCommands.FRC_1BYTE,
            [0, 2, 104, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            None,
            [SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE, 30, SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE] + [
                SensorFrcErrors.NO_FRC_RESPONSE] * 60
        ],
        [
            '1byte_temperature_count',
            SensorTypes.TEMPERATURE,
            0,
            SensorFrcCommands.FRC_1BYTE,
            [0, 2, 104, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            5,
            [SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE, 30, SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE] + [
                SensorFrcErrors.NO_FRC_RESPONSE] * 2
        ],
        [
            '2byte_temperature_no_count',
            SensorTypes.TEMPERATURE,
            0,
            SensorFrcCommands.FRC_2BYTES,
            [0, 0, 2, 0, 216, 129, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            None,
            [SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE, 29.5, SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE] + [
                SensorFrcErrors.NO_FRC_RESPONSE] * 28
        ],
        [
            '2byte_temperature_count',
            SensorTypes.TEMPERATURE,
            0,
            SensorFrcCommands.FRC_2BYTES,
            [0, 0, 2, 0, 216, 129, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            4,
            [SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE, 29.5, SensorFrcErrors.SENSOR_ERROR_OR_OUT_OF_RANGE] + [
                SensorFrcErrors.NO_FRC_RESPONSE] * 1
        ],
        [
            '4byte_bin30_no_count',
            SensorTypes.BINARY_DATA30,
            0,
            SensorFrcCommands.FRC_4BYTES,
            [0, 0, 0, 0, 1, 0, 0, 0, 40, 124, 17, 129, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            None,
            [SensorFrcErrors.FRC_NOT_IMPLEMENTED, 2165406756, SensorFrcErrors.FRC_NOT_IMPLEMENTED] + [
                SensorFrcErrors.NO_FRC_RESPONSE] * 12
        ],
        [
            '4byte_bin30_count',
            SensorTypes.BINARY_DATA30,
            0,
            SensorFrcCommands.FRC_4BYTES,
            [0, 0, 0, 0, 1, 0, 0, 0, 40, 124, 17, 129, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
             0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0],
            6,
            [SensorFrcErrors.FRC_NOT_IMPLEMENTED, 2165406756, SensorFrcErrors.FRC_NOT_IMPLEMENTED] + [
                SensorFrcErrors.NO_FRC_RESPONSE] * 3
        ]
    ])
    def test_frc_data(self, _, sensor_type: int, sensor_index: int, frc_command: int, frc_data: List[int],
                      extra_result: List[int], count: Union[int, None],
                      values: List[Union[int, float, List[int], SensorFrcErrors]]
                      ):
        sensor_data = SensorParser.frc_dpa(
            sensor_type=sensor_type,
            sensor_index=sensor_index,
            frc_command=frc_command,
            data=frc_data,
            extra_result=extra_result,
            count=count
        )
        for i in range(len(values)):
            self.assertEqual(sensor_data[i].value, values[i])

    @parameterized.expand([
        [
            'unknown_sensor',
            30,
            0,
            SensorFrcCommands.FRC_1BYTE,
            [],
            [],
            None
        ],
        [
            '2bit_too_short',
            129,
            0,
            SensorFrcCommands.FRC_2BITS,
            [],
            [],
            None
        ],
        [
            '2byte_invalid_len_no_count',
            1,
            0,
            SensorFrcCommands.FRC_2BYTES,
            [0] * 55,
            [],
            None
        ],
        [
            '2byte_invalid_len_count',
            1,
            0,
            SensorFrcCommands.FRC_2BYTES,
            [0] * 5,
            [],
            10
        ],
        [
            '4byte_invalid_len_no_count',
            160,
            0,
            SensorFrcCommands.FRC_4BYTES,
            [0] * 55,
            [],
            None
        ],
        [
            '4byte_invalid_len_count',
            160,
            0,
            SensorFrcCommands.FRC_4BYTES,
            [0] * 10,
            [0],
            3
        ]
    ])
    def test_frc_data_invalid(self, _, sensor_type: int, sensor_index: int, frc_command: int, frc_data: List[int],
                              extra_result: List[int], count: Union[int, None]):
        with self.assertRaises(ValueError):
            SensorParser.frc_dpa(
                sensor_type=sensor_type,
                sensor_index=sensor_index,
                frc_command=frc_command,
                data=frc_data,
                extra_result=extra_result,
                count=count
            )



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import FrcResponseCommands
from iqrfpy.enums.message_types import FrcMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.frc.responses.set_frc_params import SetFrcParamsResponse
from iqrfpy.peripherals.frc.frc_params import FrcParams
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': FrcMessages.SET_PARAMS,
    'msgid': 'setFrcParamsTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 64,
    'result': {
        'frcResponseTime': 0,
    },
    'dpa': b'\x00\x00\x0d\x83\x00\x00\x00\x40\x00'
}

data_ok_1: dict = {
    'mtype': FrcMessages.SET_PARAMS,
    'msgid': 'setFrcParamsTest',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'result': {
        'frcResponseTime': 24,
    },
    'dpa': b'\x00\x00\x0d\x83\x02\x04\x00\x23\x18'
}

data_error: dict = {
    'mtype': FrcMessages.SET_PARAMS,
    'msgid': 'setFrcParamsTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x0d\x83\x04\x04\x01\x23'
}


class SetDpaParamsResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, SetFrcParamsResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, SetFrcParamsResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, SetFrcParamsResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, SetFrcParamsResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, SetFrcParamsResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, SetFrcParamsResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.FRC)
        with self.subTest():
            self.assertEqual(response.pcmd, FrcResponseCommands.SET_PARAMS)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, FrcMessages.SET_PARAMS)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            SetFrcParamsResponse.from_dpa(b'\x00\x00\x0d\x83\x00\x00\x00\x40\x21\x05\x33')

    @parameterized.expand([
        [
            'from_dpa',
            FrcParams.from_int(data_ok['result']['frcResponseTime']),
            SetFrcParamsResponse.from_dpa(data_ok['dpa']),
        ],
        [
            'from_dpa',
            FrcParams.from_int(data_ok_1['result']['frcResponseTime']),
            SetFrcParamsResponse.from_dpa(data_ok_1['dpa']),
        ],
        [
            'from_json',
            FrcParams.from_int(data_ok['result']['frcResponseTime']),
            SetFrcParamsResponse.from_json(generate_json_response(data_ok)),
        ],
        [
            'from_json',
            FrcParams.from_int(data_ok_1['result']['frcResponseTime']),
            SetFrcParamsResponse.from_json(generate_json_response(data_ok_1)),
        ]
    ])
    def test_get_frc_params(self, _, frc_params: FrcParams, response: SetFrcParamsResponse):
        response_frc_params = response.frc_params
        with self.subTest():
            self.assertEqual(response_frc_params.offline_frc, frc_params.offline_frc)
        with self.subTest():
            self.assertEqual(response_frc_params.frc_response_time, frc_params.frc_response_time)

    @parameterized.expand([
        ['from_dpa_error', SetFrcParamsResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', SetFrcParamsResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_frc_params_error(self, _, response: SetFrcParamsResponse):
        self.assertIsNone(response.frc_params)



"""Generic request message."""

import math
from typing import List, Optional, Union
from iqrfpy.enums.commands import Command
from iqrfpy.enums.message_types import GenericMessages
from iqrfpy.enums.peripherals import Peripheral
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.irequest import IRequest

__all__ = ['GenericRequest']


class GenericRequest(IRequest):
    """Generic request class."""

    def __init__(self, nadr: int, pnum: Union[Peripheral, int], pcmd: Union[Command, int],
                 hwpid: int = dpa_constants.HWPID_MAX, pdata: Optional[List[int]] = None,
                 dpa_rsp_time: Optional[float] = None, dev_process_time: Optional[float] = None,
                 msgid: Optional[str] = None):
        """Generic request constructor.

        Args:
            nadr (int): Device address.
            pnum (Union[Peripheral, int]): Peripheral number.
            pcmd (Union[Command, int]): Peripheral command.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535 (Ignore HWPID check).
            pdata (List[int], optional): Request data.
            dpa_rsp_time (float, optional): DPA request timeout in seconds. Defaults to None.
            dev_process_time (float, optional): Device processing time. Defaults to None.
            msgid (str, optional): JSON API message ID. Defaults to None. If the parameter is not specified, a random
                UUIDv4 string is generated and used.
        """
        super().__init__(
            nadr=nadr,
            pnum=pnum,
            pcmd=pcmd,
            m_type=GenericMessages.RAW,
            hwpid=hwpid,
            pdata=pdata,
            dpa_rsp_time=dpa_rsp_time,
            dev_process_time=dev_process_time,
            msgid=msgid
        )

    def to_dpa(self, mutable: bool = False) -> Union[bytes, bytearray]:
        """DPA request serialization method.

        Args:
            mutable (bool, optional): Serialize into mutable byte representation of DPA request packet.
                Defaults to False.

        Returns:
            :obj:`bytes`: Immutable byte representation of DPA request packet.\n
            :obj:`bytearray`: Mutable byte representation of DPA request packet (if argument mutable is True).
        """
        return super().to_dpa(mutable=mutable)

    def to_json(self) -> dict:
        """JSON API request serialization method.

        Returns:
            :obj:`dict`: JSON API request object.
        """
        dpa: List[int] = [self._nadr, 0, self._pnum, self._pcmd, self._hwpid & 0xFF, (self._hwpid >> 8) & 0xFF]
        if self._pdata is not None:
            dpa.extend(self._pdata)
        json: dict = {
            'mType': self._mtype.value,
            'data': {
                'msgId': self._msgid,
                'req': {
                    'rData': '.'.join([f'{x:02x}' for x in dpa])
                },
                'returnVerbose': True,
            },
        }
        if self._dpa_rsp_time is not None:
            json['data']['timeout'] = math.ceil(self._dpa_rsp_time * 1000)
        return json



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import OSResponseCommands
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.exceptions import DpaResponsePacketLengthError
from iqrfpy.peripherals.os.responses.read import ReadResponse, OsReadData
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': OSMessages.READ,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 71,
    'result': {
        'mid': 2165406807,
        'osVersion': 70,
        'trMcuType': 36,
        'osBuild': 2264,
        'rssi': 0,
        'supplyVoltage': 3.0013793103448276,
        'flags': 32,
        'slotLimits': 117,
        'ibk': [
           228,
           130,
           127,
           38,
           186,
           44,
           106,
           194,
           118,
           84,
           88,
           173,
           52,
           27,
           86,
           46
        ],
        'dpaVer': 1047,
        'perNr': 0,
        'embeddedPers': [
           0,
           2,
           3,
           4,
           5,
           6,
           7,
           9,
           10,
           13
        ],
        'hwpid': 0,
        'hwpidVer': 0,
        'flagsEnum': 5,
        'userPer': []
    },
    'dpa': b'\x00\x00\x02\x80\x00\x00\x00\x47\x57\x7c\x11\x81\x46\x24\xd8\x08\x00\x28\x20\x75\xe4\x82\x7f\x26\xba\x2c'
           b'\x6a\xc2\x76\x54\x58\xad\x34\x1b\x56\x2e\x17\x04\x00\xfd\x26\x00\x00\x00\x00\x00\x00\x05'
}

data_error: dict = {
    'mtype': OSMessages.READ,
    'msgid': 'readTest',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 7,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x02\x80\x04\x04\x07\x23'
}


class ReadResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, ReadResponse.from_dpa(data_ok['dpa']), False],
        ['from_json', data_ok, ReadResponse.from_json(generate_json_response(data_ok)), True],
        ['from_dpa_error', data_error, ReadResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, ReadResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.OS)
        with self.subTest():
            self.assertEqual(response.pcmd, OSResponseCommands.READ)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, OSMessages.READ)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(DpaResponsePacketLengthError):
            ReadResponse.from_dpa(b'\x00\x00\x02\x80\x00\x00\x00\x47')

    @parameterized.expand([
        ['from_dpa', OsReadData(data_ok['result']), ReadResponse.from_dpa(data_ok['dpa'])],
        ['from_json', OsReadData(data_ok['result']), ReadResponse.from_json(generate_json_response(data_ok))]
    ])
    def test_get_os_read_data(self, _, data: OsReadData, response: ReadResponse):
        self.assertEqual(response.os_read_data, data)

    @parameterized.expand([
        ['from_dpa_error', ReadResponse.from_dpa(data_error['dpa'])],
        ['from_json_error', ReadResponse.from_json(generate_json_response(data_error))]
    ])
    def test_get_os_read_data_error(self, _, response: ReadResponse):
        self.assertIsNone(response.os_read_data)



"""UART Clear Write Read response message."""

from typing import List, Optional
from iqrfpy.iresponse import IResponseGetterMixin
from iqrfpy.enums.commands import UartResponseCommands
from iqrfpy.enums.message_types import UartMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants
from iqrfpy.utils.dpa import ResponsePacketMembers, ResponseCodes
from iqrfpy.utils.validators import DpaValidator, JsonValidator


class ClearWriteReadResponse(IResponseGetterMixin):
    """UART Clear Write Read response class."""

    __slots__ = ('_read_data',)

    def __init__(self, nadr: int, hwpid: int = dpa_constants.HWPID_MAX, rcode: int = 0, dpa_value: int = 0,
                 msgid: Optional[str] = None, pdata: Optional[List[int]] = None, result: Optional[dict] = None):
        """Clear Write Read response constructor.

        Args:
            nadr (int): Device address.
            hwpid (int, optional): Hardware profile ID. Defaults to 65535, this value ignores HWPID check.
            rcode (int, optional): Response code. Defaults to 128.
            dpa_value (int, optional): DPA value. Defaults to 0.
            pdata (List[int], optional): DPA response data. Defaults to None.
            msgid (str, optional): Message ID. Defaults to None.
            result (dict, optional): JSON response data. Defaults to None.
        """
        super().__init__(
            nadr=nadr,
            pnum=EmbedPeripherals.UART,
            pcmd=UartResponseCommands.CLEAR_WRITE_READ,
            m_type=UartMessages.CLEAR_WRITE_READ,
            hwpid=hwpid,
            rcode=rcode,
            dpa_value=dpa_value,
            msgid=msgid,
            pdata=pdata,
            result=result
        )
        self._read_data = result['readData'] if rcode == ResponseCodes.OK else None

    @property
    def read_data(self) -> Optional[List[int]]:
        """:obj:`list` of :obj:`int` or :obj:`None`: Data read from RX buffer.

        Getter only.
        """
        return self._read_data

    @classmethod
    def from_dpa(cls, dpa: bytes) -> 'ClearWriteReadResponse':
        """DPA response factory method.

        Parses DPA data and constructs :obj:`ClearWriteReadResponse` object.

        Args:
            dpa (bytes): DPA response bytes.
        Returns:
            :obj:`ClearWriteReadResponse`: Response message object.
        """
        DpaValidator.base_response_length(dpa=dpa)
        nadr = dpa[ResponsePacketMembers.NADR]
        hwpid = Common.hwpid_from_dpa(dpa[ResponsePacketMembers.HWPID_HI], dpa[ResponsePacketMembers.HWPID_LO])
        rcode = dpa[ResponsePacketMembers.RCODE]
        dpa_value = dpa[ResponsePacketMembers.DPA_VALUE]
        pdata = None
        result = None
        if rcode == ResponseCodes.OK:
            pdata = Common.pdata_from_dpa(dpa=dpa)
            result = {'readData': pdata}
        return cls(nadr=nadr, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)

    @classmethod
    def from_json(cls, json: dict) -> 'ClearWriteReadResponse':
        """JSON response factory method.

        Parses JSON API response and constructs :obj:`ClearWriteReadResponse` object.

        Args:
            json (dict): JSON API response.
        Returns:
            :obj:`ClearWriteReadResponse`: Response message object.
        """
        JsonValidator.response_received(json=json)
        nadr = Common.nadr_from_json(json=json)
        msgid = Common.msgid_from_json(json=json)
        hwpid = Common.hwpid_from_json(json=json)
        dpa_value = Common.dpa_value_from_json(json=json)
        rcode = Common.rcode_from_json(json=json)
        pdata = Common.pdata_from_json(json=json)
        result = Common.result_from_json(json=json) if rcode == ResponseCodes.OK else None
        return cls(nadr=nadr, msgid=msgid, hwpid=hwpid, rcode=rcode, dpa_value=dpa_value, pdata=pdata, result=result)



"""Sensor Written Data parameters."""

from typing import List
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.utils.common import Common
import iqrfpy.utils.dpa as dpa_constants


class SensorWrittenData:
    """Sensor Written Data class."""

    __slots__ = '_index', '_data'

    def __init__(self, index: int, data: List[int]):
        """Sensor Written Data constructor.

        Args:
            index (int): Sensor index.
            data (List[int]): Data to write.
        """
        self._validate(index=index, data=data)
        self._index = index
        self._data = data

    def _validate(self, index: int, data: List[int]):
        """Validate parameters.

        Args:
            index (int): Sensor index.
            data (List[int]): Data to write.
        """
        self._validate_index(index)
        self._validate_data(data)

    @staticmethod
    def _validate_index(index: int):
        """Validate sensor index parameter.

        Args:
            index (int): Sensor index.
        Raises:
            RequestParameterInvalidValueError: If index is less than 0 or greater than 31.
        """
        if not dpa_constants.SENSOR_INDEX_MIN <= index <= dpa_constants.SENSOR_INDEX_MAX:
            raise RequestParameterInvalidValueError('Index value should be between 0 and 31.')

    @property
    def index(self) -> int:
        """:obj:`int`: Sensor index.

        Getter and setter.
        """
        return self._index

    @index.setter
    def index(self, value: int):
        self._validate_index(value)
        self._index = value

    @staticmethod
    def _validate_data(data: List[int]):
        """Validate data to write.

        Args:
            data (List[int]): Data to write.
        Raises:
            RequestParameterInvalidValueError: If data contains values not in range from 0 to 255.
        """
        if not Common.values_in_byte_range(data):
            raise RequestParameterInvalidValueError('Data values should be between 0 and 255.')

    @property
    def data(self) -> List[int]:
        """:obj:`list` of :obj:`int`: Data to write.

        Getter and setter.
        """
        return self._data

    @data.setter
    def data(self, value: List[int]):
        self._validate_data(value)
        self._data = value

    def to_pdata(self) -> List[int]:
        """Serialize index and data to parameters.

        Returns:
            :obj:`list` of `int`: Serialized written data.
        """
        return [self.index] + self.data



import unittest
from parameterized import parameterized
from iqrfpy.exceptions import RequestParameterInvalidValueError
from iqrfpy.peripherals.ram.requests.read_any import ReadAnyRequest


class ReadAnyRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x05\x00\x05\x0f\xff\xff\x0a\x00\x05'

    @parameterized.expand([
        [5, 10, 5, b'\x05\x00\x05\x0f\xff\xff\x0a\x00\x05'],
        [0, 5, 22, b'\x00\x00\x05\x0f\xff\xff\x05\x00\x16'],
        [5, 65535, 10, b'\x05\x00\x05\x0f\xff\xff\xff\xff\x0a']
    ])
    def test_to_dpa(self, nadr: int, address: int, length: int, expected):
        request = ReadAnyRequest(nadr=nadr, address=address, length=length)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        [5, 10, 5],
        [0, 5, 22],
        [5, 12345, 20]
    ])
    def test_to_json(self, nadr: int, address: int, length: int):
        request = ReadAnyRequest(nadr=nadr, address=address, length=length, msgid='readTest')
        with self.assertRaises(NotImplementedError):
            request.to_json()

    @parameterized.expand([
        [2, b'\x05\x00\x05\x0f\xff\xff\x02\x00\x05'],
        [17, b'\x05\x00\x05\x0f\xff\xff\x11\x00\x05'],
        [65534, b'\x05\x00\x05\x0f\xff\xff\xfe\xff\x05'],
    ])
    def test_set_address(self, address, dpa):
        request = ReadAnyRequest(nadr=5, address=10, length=5, msgid='readTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.address = address
        self.assertEqual(
            request.to_dpa(),
            dpa
        )

    @parameterized.expand([
        [20, b'\x05\x00\x05\x0f\xff\xff\x0a\x00\x14'],
        [13, b'\x05\x00\x05\x0f\xff\xff\x0a\x00\x0d'],
    ])
    def test_set_length(self, length, dpa):
        request = ReadAnyRequest(nadr=5, address=10, length=5, msgid='readTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        request.length = length
        self.assertEqual(
            request.to_dpa(),
            dpa
        )

    @parameterized.expand([
        [-1],
        [65536],
        [100000]
    ])
    def test_set_address_invalid(self, address):
        request = ReadAnyRequest(nadr=5, address=10, length=5, msgid='readTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.address = address

    @parameterized.expand([
        [-1],
        [256],
        [1000]
    ])
    def test_set_length_invalid(self, length):
        request = ReadAnyRequest(nadr=5, address=10, length=5, msgid='readTest')
        with self.assertRaises(RequestParameterInvalidValueError):
            request.length = length

    @parameterized.expand([
        [-1, 1],
        [65536, 10],
        [100000, 15],
        [10, -1],
        [10, 256]
    ])
    def test_construct_invalid(self, address, length):
        with self.assertRaises(RequestParameterInvalidValueError):
            ReadAnyRequest(nadr=0, address=address, length=length)



import unittest
from parameterized import parameterized
from iqrfpy.enums.commands import CoordinatorResponseCommands
from iqrfpy.enums.message_types import CoordinatorMessages
from iqrfpy.enums.peripherals import EmbedPeripherals
from iqrfpy.peripherals.coordinator.responses.set_mid import SetMidResponse
from tests.helpers.json import generate_json_response

data_ok: dict = {
    'mtype': CoordinatorMessages.SET_MID,
    'msgid': 'setMidResponse',
    'nadr': 0,
    'hwpid': 0,
    'rcode': 0,
    'dpa_value': 64,
    'dpa': b'\x00\x00\x00\x83\x00\x00\x00\x40'
}

data_ok_1: dict = {
    'mtype': CoordinatorMessages.SET_MID,
    'msgid': 'setMidResponse',
    'nadr': 0,
    'hwpid': 1026,
    'rcode': 0,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x83\x02\x04\x00\x23'
}

data_error: dict = {
    'mtype': CoordinatorMessages.SET_MID,
    'msgid': 'setMidResponse',
    'nadr': 0,
    'hwpid': 1028,
    'rcode': 1,
    'dpa_value': 35,
    'dpa': b'\x00\x00\x00\x83\x04\x04\x01\x23'
}


class SetMIDResponseTestCase(unittest.TestCase):

    @parameterized.expand([
        ['from_dpa', data_ok, SetMidResponse.from_dpa(data_ok['dpa']), False],
        ['from_dpa', data_ok_1, SetMidResponse.from_dpa(data_ok_1['dpa']), False],
        ['from_json', data_ok, SetMidResponse.from_json(generate_json_response(data_ok)), True],
        ['from_json', data_ok_1, SetMidResponse.from_json(generate_json_response(data_ok_1)), True],
        ['from_dpa_error', data_error, SetMidResponse.from_dpa(data_error['dpa']), False],
        ['from_json_error', data_error, SetMidResponse.from_json(generate_json_response(data_error)), True]
    ])
    def test_factory_methods_ok(self, _, response_data, response, json):
        with self.subTest():
            self.assertEqual(response.nadr, response_data['nadr'])
        with self.subTest():
            self.assertEqual(response.pnum, EmbedPeripherals.COORDINATOR)
        with self.subTest():
            self.assertEqual(response.pcmd, CoordinatorResponseCommands.SET_MID)
        with self.subTest():
            self.assertEqual(response.hwpid, response_data['hwpid'])
        with self.subTest():
            self.assertEqual(response.rcode, response_data['rcode'])
        if json:
            with self.subTest():
                self.assertEqual(response.mtype, CoordinatorMessages.SET_MID)
            with self.subTest():
                self.assertEqual(response.msgid, response_data['msgid'])

    def test_from_dpa_invalid(self):
        with self.assertRaises(ValueError):
            SetMidResponse.from_dpa(b'\x00\x00\x00\x93\x00\x00\x00\x22\x01')



import unittest
from parameterized import parameterized
from iqrfpy.peripherals.coordinator.requests.discovery import DiscoveryRequest


class DiscoveryRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.dpa = b'\x00\x00\x00\x07\xff\xff\x06\x00'
        self.json = {
            'mType': 'iqrfEmbedCoordinator_Discovery',
            'data': {
                'msgId': 'discoveryTest',
                'req': {
                    'nAdr': 0,
                    'hwpId': 65535,
                    'param': {
                        'txPower': 6,
                        'maxAddr': 0
                    }
                },
                'returnVerbose': True
            }
        }

    @parameterized.expand([
        ['tx_power', 0, 0, b'\x00\x00\x00\x07\xff\xff\x00\x00'],
        ['tx_power', 7, 0, b'\x00\x00\x00\x07\xff\xff\x07\x00'],
        ['max_addr', 7, 5, b'\x00\x00\x00\x07\xff\xff\x07\x05'],
        ['max_addr', 7, 239, b'\x00\x00\x00\x07\xff\xff\x07\xef']
    ])
    def test_to_dpa(self, _, tx_power, max_addr, expected):
        request = DiscoveryRequest(tx_power=tx_power, max_addr=max_addr)
        self.assertEqual(
            request.to_dpa(),
            expected
        )

    @parameterized.expand([
        ['tx_power', 0, 0],
        ['tx_power', 7, 0],
        ['max_addr', 7, 5],
        ['max_addr', 7, 239]
    ])
    def test_to_json(self, _, tx_power, max_addr):
        request = DiscoveryRequest(tx_power=tx_power, max_addr=max_addr, msgid='discoveryTest')
        self.json['data']['req']['param']['txPower'] = tx_power
        self.json['data']['req']['param']['maxAddr'] = max_addr
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [5, b'\x00\x00\x00\x07\xff\xff\x05\x00'],
        [7, b'\x00\x00\x00\x07\xff\xff\x07\x00']
    ])
    def test_set_tx_power(self, tx_power, dpa):
        request = DiscoveryRequest(tx_power=6, max_addr=0, msgid='discoveryTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.tx_power = tx_power
        self.json['data']['req']['param']['txPower'] = tx_power
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [0, b'\x00\x00\x00\x07\xff\xff\x06\x00'],
        [239, b'\x00\x00\x00\x07\xff\xff\x06\xef']
    ])
    def test_set_max_addr(self, max_addr, dpa):
        request = DiscoveryRequest(tx_power=6, max_addr=0, msgid='discoveryTest')
        self.assertEqual(
            request.to_dpa(),
            self.dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )
        request.max_addr = max_addr
        self.json['data']['req']['param']['maxAddr'] = max_addr
        self.assertEqual(
            request.to_dpa(),
            dpa
        )
        self.assertEqual(
            request.to_json(),
            self.json
        )

    @parameterized.expand([
        [-1, 0],
        [256, 0],
        [7, -10],
        [7, 520]
    ])
    def test_construct_invalid(self, tx_power, max_addr):
        with self.assertRaises(ValueError):
            DiscoveryRequest(tx_power=tx_power, max_addr=max_addr)



import unittest
from iqrfpy.enums.message_types import OSMessages
from iqrfpy.peripherals.os.requests.read_tr_conf import ReadTrConfRequest


class ReadTrConfRequestTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.request = ReadTrConfRequest(nadr=2, msgid='readTrConfTest')
        self.dpa = b'\x02\x00\x02\x02\xff\xff'
        self.json = {
            'mType': OSMessages.READ_CFG.value,
            'data': {
                'msgId': 'readTrConfTest',
                'req': {
                    'nAdr': 2,
                    'hwpId': 65535,
                    'param': {},
                },
                'returnVerbose': True
            },
        }

    def test_to_dpa(self):
        self.assertEqual(
            self.request.to_dpa(),
            self.dpa
        )

    def test_to_json(self):
        self.assertEqual(
            self.request.to_json(),
            self.json
        )



"""OS Sleep request parameters."""

from typing import List
from iqrfpy.exceptions import RequestParameterInvalidValueError
import iqrfpy.utils.dpa as dpa_constants


class OsSleepParams:
    """OS Sleep Params class."""

    __slots__ = '_time', 'wake_up_on_negative_edge', 'calibrate_before_sleep', 'flash_led_after_sleep', \
        'wake_up_on_positive_edge', 'use_milliseconds', 'use_deep_sleep'

    def __init__(self, time: int = 0, wake_up_on_negative_edge: bool = False, calibrate_before_sleep: bool = False,
                 flash_led_after_sleep: bool = False, wake_up_on_positive_edge: bool = False,
                 use_milliseconds: bool = False, use_deep_sleep: bool = False):
        """Sleep Params constructor.

        Args:
            time (int): Sleep time.
            wake_up_on_negative_edge (bool): Wake up on PORT B, bit 4 negative edge change.
            calibrate_before_sleep (bool): Run calibration process before sleep.
            flash_led_after_sleep (bool): Flash green LED once after waking up.
            wake_up_on_positive_edge (bool): Wake up on PORT B, bit 4 positive edge change.
            use_milliseconds (bool): Use sleep time with unit of 32.768 ms instead of 2.097s.
            use_deep_sleep (bool): The IC is shutdown during sleep.
        """
        self._validate_time(time)
        self._time = time
        self.wake_up_on_negative_edge = wake_up_on_negative_edge
        """Wake up on PORT B, bit 4 negative edge change."""
        self.calibrate_before_sleep = calibrate_before_sleep
        """Run calibration process before sleep."""
        self.flash_led_after_sleep = flash_led_after_sleep
        """Flash green LED once after waking up."""
        self.wake_up_on_positive_edge = wake_up_on_positive_edge
        """Wake up on PORT B, bit 4 positive edge change."""
        self.use_milliseconds = use_milliseconds
        """Use sleep time with unit of 32.768 ms instead of 2.097s."""
        self.use_deep_sleep = use_deep_sleep
        """The IC is shutdown during sleep."""

    @staticmethod
    def _validate_time(time: int):
        """Validate sleep time parameter.

        Args:
            time (int): Sleep time.
        Raises:
            RequestParameterInvalidValueError: If time is less than 0 or greater than 255.
        """
        if not dpa_constants.WORD_MIN <= time <= dpa_constants.WORD_MAX:
            raise RequestParameterInvalidValueError('Time value should be between 0 and 65535.')

    @property
    def time(self) -> int:
        """:obj:`int`: Sleep time.

        Getter and setter.
        """
        return self._time

    @time.setter
    def time(self, value: int):
        self._validate_time(value)
        self._time = value

    def _calculate_control(self) -> int:
        """Convert flags into a single value.

        Returns:
            :obj:`int`: Flags value.
        """
        return self.wake_up_on_negative_edge | (self.calibrate_before_sleep << 1) | \
            (self.flash_led_after_sleep << 2) | (self.wake_up_on_positive_edge << 3) | \
            (self.use_milliseconds << 4) | (self.use_deep_sleep << 5)

    def to_pdata(self) -> List[int]:
        """Serialize sleep parameters into DPA request pdata.

        Returns:
            :obj:`list` of :obj:`int`: Serialized DPA request pdata.
        """
        return [self._time & 0xFF, (self.time >> 8) & 0xFF, self._calculate_control()]

    def to_json(self) -> dict:
        """Serializes sleep parameters into JSON API request data.

        Returns:
            :obj:`dict`: Serialized JSON API request data.
        """
        return {
            'time': self._time,
            'control': self._calculate_control()
        }



import json
import random
import string
import threading

from typing import Callable, List, Optional
from paho.mqtt.client import Client
from iqrfpy.enums.message_types import MessageType
from iqrfpy.exceptions import TransportNotConnectedError, MessageNotReceivedError, DpaRequestTimeoutError, \
    JsonRequestTimeoutError, UnsupportedMessageTypeError, JsonResponseNotSupportedError
from iqrfpy.response_factory import ResponseFactory
from iqrfpy.transports.itransport import ITransport
from iqrfpy.irequest import IRequest
from iqrfpy.iresponse import IResponse
from iqrfpy.confirmation import Confirmation

__all__ = [
    'MqttTransportParams',
    'MqttTransport',
]


class MqttParamsError(Exception):
    """MQTT parameters error.

    Raised if MQTT parameter does not validate.
    """

    pass


class MqttTransportParams:
    """MQTT transport parameters class."""

    __slots__ = '_host', '_port', '_client_id', '_user', '_password', '_request_topic', '_response_topic', '_qos', \
        '_keepalive'

    def __init__(self, host: str = 'localhost', port: int = 1883, client_id: Optional[str] = None,
                 user: Optional[str] = None, password: Optional[str] = None, request_topic: Optional[str] = None,
                 response_topic: Optional[str] = None, qos: int = 1, keepalive: int = 60):
        """MQTT transport parameters constructor.

        Args:
            host (str): Broker hostname. (Defaults to 'localhost')
            port (int): Broker port. (Defaults to 1883)
            client_id (str, optional): Client ID. If no client ID is specified, a randomly generated one will be used.
            user (str, optional): Broker username.
            password (str, optional): Broker password.
            request_topic (str, optional): Topic to publish requests to.
            response_topic (str, optional): Topic to subscribe to for responses.
            qos (int): Quality of service. (Defaults to 1)
            keepalive (int): Keep alive interval. (Defaults to 60)
        """
        self._validate(port=port, qos=qos)
        self._host = host
        self._port = port
        self._client_id = client_id if client_id is not None else \
            ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(16))
        self._user = user
        self._password = password
        self._request_topic = request_topic
        self._response_topic = response_topic
        self._qos = qos
        self._keepalive = keepalive

    def _validate(self, port: int, qos: int):
        """Validate transport parameters.

        Args:
            port (int): Port.
            qos (int): Quality of service.
        """
        self._validate_port(port)
        self._validate_qos(qos)

    @property
    def host(self) -> str:
        """:obj:`str`: Broker hostname.

        Getter and setter.
        """
        return self._host

    @host.setter
    def host(self, value: str):
        self._host = value

    @staticmethod
    def _validate_port(port: int):
        """Validate broker port number.

        Args:
            port (int): Broker port number.
        Raises:
            MqttParamsError: If port is less than 1024 and greater than 65535.
        """
        if not (1024 <= port <= 65535):
            raise MqttParamsError('Port value should be between 1024 and 65535.')

    @property
    def port(self) -> int:
        """:obj:`int`: Broker port.

        Getter and setter.
        """
        return self._port

    @port.setter
    def port(self, value: int):
        self._validate_port(value)
        self._port = value

    @property
    def client_id(self) -> str:
        """:obj:`str`: Client ID.

        Getter and setter.
        """
        return self._client_id

    @client_id.setter
    def client_id(self, value: str):
        self._client_id = value

    @property
    def user(self) -> Optional[str]:
        """:obj:`str` or :obj:`None`: Broker username.

        Getter and setter.
        """
        return self._user

    @user.setter
    def user(self, value: Optional[str]):
        self._user = value

    @property
    def password(self) -> Optional[str]:
        """:obj:`str` or :obj:`None`: Broker password.

        Getter and setter.
        """
        return self._password

    @password.setter
    def password(self, value: Optional[str]):
        self._password = value

    @property
    def request_topic(self) -> Optional[str]:
        """:obj:`str` or :obj:`None`: Topic to publish requests to.

        Getter and setter.
        """
        return self._request_topic

    @request_topic.setter
    def request_topic(self, value: str):
        self._request_topic = value

    @property
    def response_topic(self) -> Optional[str]:
        """:obj:`str` or :obj:`None`: Topic to subscribe to for responses.

        Getter and setter.
        """
        return self._response_topic

    @response_topic.setter
    def response_topic(self, value: str):
        self._response_topic = value

    @staticmethod
    def _validate_qos(qos: int):
        """Validate quality of service parameter.

        Args:
            qos (int): Quality of service.
        Raises:
            MqttParamsError: If qos is less than 0 or greater than 2.
        """
        if not (0 <= qos <= 2):
            raise MqttParamsError('QoS value should be between 0 and 2.')

    @property
    def qos(self) -> int:
        """:obj:`int`: Quality of service.

        Getter and setter.
        """
        return self._qos

    @qos.setter
    def qos(self, value: int):
        self._validate_qos(value)
        self._qos = value

    @property
    def keepalive(self) -> int:
        """:obj:`int`: Keep alive interval.

        Getter and setter.
        """
        return self._keepalive

    @keepalive.setter
    def keepalive(self, value: int):
        self._keepalive = value


class MqttTransport(ITransport):
    """MQTT transport class."""

    __slots__ = '_client', '_params', '_callback', '_global_request_timeout', '_cv', '_msg_ids', '_m_type', \
        '_response', '_dpa_timeout', '_received_timeout', '_unsupported'

    def __init__(self, params: MqttTransportParams, callback: Optional[Callable] = None, auto_init: bool = False,
                 global_request_timeout: Optional[float] = 5):
        """MQTT transport constructor.

        Args:
            params (MqttTransportParams): MQTT parameters.
            callback (Callable, optional): Callback to process responses.
            auto_init (bool): Initialize and establish connection automatically.
            global_request_timeout (float, optional): Request timeout, used if timeout is not explicitly
                passed to receive methods.
        """
        self._client: Optional[Client] = None
        self._params: MqttTransportParams = params
        self._callback: Optional[Callable] = callback
        self._global_request_timeout: float = global_request_timeout
        self._cv: threading.Condition = threading.Condition()
        self._sub_cv: threading.Condition = threading.Condition()
        self._msg_ids: List[str] = []
        self._m_type: Optional[MessageType] = None
        self._response: Optional[IResponse] = None
        self._dpa_timeout: Optional[float] = None
        self._received_timeout: bool = False
        self._unsupported: bool = False
        if auto_init:
            self.initialize()

    def initialize(self) -> None:
        """Initialize and establish MQTT connection.

        The method initializes MQTT client, sets callbacks, connects to broker and subscribes to response topic.
        """
        self._client = Client(self._params.client_id)
        self._client.on_connect = self._connect_callback
        self._client.on_message = self._message_callback
        self._client.on_subscribe = self._subscribe_callback
        if self._params.user is not None and self._params.password is not None:
            self._client.username_pw_set(self._params.user, self._params.password)
        self._client.connect(self._params.host, self._params.port)
        self._client.loop_start()
        with self._sub_cv:
            self._sub_cv.wait()

    def terminate(self, force: bool = False) -> None:
        """Terminate MQTT connection.

        Args:
            force (bool): Force terminate connection. (Defaults to False)
        """
        self._client.disconnect()
        self._client.loop_stop(force=force)

    def _connect_callback(self, client, userdata, flags, rc):
        """On connect callback.

        Subscribes to response topic upon establishing connection.
        """
        # pylint: disable=W0613
        if rc == 0:
            self._client.subscribe(self._params.response_topic, self._params.qos)

    def _subscribe_callback(self, client, userdata, mid, granted_qos):
        """On subscribe callback.

        Notifies topic subscription conditional variable.
        """
        # pylint: disable=W0613
        with self._sub_cv:
            self._sub_cv.notify()

    def _message_callback(self, client, userdata, message):
        """On message callback.

        Processes messages published to response topic.
        If a callback has been assigned, then the response is passed to the callback as well.
        """
        # pylint: disable=W0613
        payload = json.loads(message.payload.decode('utf-8'))
        try:
            response = ResponseFactory.get_response_from_json(payload)
        except MessageNotReceivedError as err:
            if err.msgid == self._msg_ids[0]:
                self._received_timeout = True
                with self._cv:
                    self._cv.notify()
            return
        except UnsupportedMessageTypeError as err:
            if err.msgid == self._msg_ids[0]:
                self._unsupported = True
                with self._cv:
                    self._cv.notify()
            return
        if len(self._msg_ids) > 0 and response.msgid == self._msg_ids[0] and response.mtype == self._m_type:
            self._msg_ids.pop(0)
            self._response = response
            with self._cv:
                self._cv.notify()
        if self._callback is not None:
            self._callback(response)

    def is_connected(self) -> bool:
        """Check if client is connected.

        Returns:
            bool: True if client is connected to broker, False otherwise
        """
        return self._client.is_connected()

    def send(self, request: IRequest) -> None:
        """Publish message to request topic.

        Args:
            request (IRequest): Request to publish.
        Raises:
            TransportNotConnectedError: If client is not connected to the broker.
        """
        self._response = None
        self._m_type = None
        self._dpa_timeout = None
        self._received_timeout = False
        self._unsupported = False
        if not self._client.is_connected():
            raise TransportNotConnectedError(f'MQTT client {self._params.client_id} not connected to broker.')
        self._client.publish(
            topic=self._params.request_topic,
            payload=json.dumps(request.to_json()),
            qos=self._params.qos
        )
        if request.nadr == 255:
            return
        self._dpa_timeout = request.dpa_rsp_time
        self._msg_ids.append(request.msgid)
        self._m_type = request.mtype

    def confirmation(self) -> Confirmation:
        """Return confirmation message.

        Daemon JSON API does not send confirmation messages separately.

        Raises:
            NotImplementedError: Confirmation not implemented.
        """
        raise NotImplementedError('Method not implemented.')

    def receive(self, timeout: Optional[float] = None) -> Optional[IResponse]:
        """Receive message synchronously.

        Args:
            timeout (float, optional): Time to wait for request to be processed and response sent.
                If not specified, :obj:`global_request_timeout` is used.
        Raises:
            DpaRequestTimeoutError: If DPA request timed out.
            JsonRequestTimeoutError: If JSON API response was received, but DPA request timed out, or if the JSON API
                response was not received within specified timeout.
            JsonResponseNotSupportedError: If JSON API response with expected message ID was received,
                but the response message type is not supported.
        """
        if len(self._msg_ids) == 0:
            return None
        timeout_to_use = timeout if timeout is not None else self._global_request_timeout
        with self._cv:
            self._cv.wait(timeout=timeout_to_use)
        if self._response is None:
            msg_id = self._msg_ids.pop(0)
            if self._received_timeout and self._dpa_timeout is not None:
                self._received_timeout = False
                raise DpaRequestTimeoutError(f'DPA request timed out (timeout {self._dpa_timeout} seconds).')
            else:
                if self._received_timeout:
                    self._received_timeout = False
                    raise JsonRequestTimeoutError(
                        f'Response message to request with ID {msg_id} received, but DPA request timed out.'
                    )
                elif self._unsupported:
                    self._unsupported = False
                    raise JsonResponseNotSupportedError(
                        f'Response message to request with ID {msg_id} received, but is not supported.'
                    )
                else:
                    raise JsonRequestTimeoutError(
                        f'Response message to request with ID {msg_id} not received within the specified time of '
                        f'{timeout_to_use} seconds.')
        return self._response

    def send_and_receive(self, request: IRequest, timeout: Optional[float] = None) -> Optional[IResponse]:
        """Publish request to request topic and wait for response.

        Args:
            request (IRequest): Request to publish.
            timeout (float, optional): Time to wait for request to be processed and response sent.
                If not specified, :obj:`global_request_timeout` is used.
        """
        self.send(request)
        if request.nadr == 255:
            return None
        return self.receive(timeout)

    def set_receive_callback(self, callback: Callable[[IResponse], None]) -> None:
        """Assign callback for messages published to the response topic.

        Args:
            callback (Callable[[IResponse], None]): Callback used to process messages published to response topic.
        """
        self._callback = callback



"""OS peripheral response messages."""

from .read import ReadResponse
from .reset import ResetResponse
from .restart import RestartResponse
from .read_tr_conf import ReadTrConfResponse, OsTrConfData
from .write_tr_conf import WriteTrConfResponse
from .rfpgm import RfpgmResponse
from .sleep import SleepResponse
from .set_security import SetSecurityResponse

__all__ = [
    'ReadResponse',
    'ResetResponse',
    'RestartResponse',
    'ReadTrConfResponse',
    'OsTrConfData',
    'WriteTrConfResponse',
    'RfpgmResponse',
    'SleepResponse',
    'SetSecurityResponse'
]



"""Utilities module.

This module contains auxiliary methods for handling and validation of DPA and JSON API responses and parsing of sensor data.
It also contains DPA and Sensor standard constants and methods for conversion of data between various types and formats.
"""



"""LEDG peripheral response messages."""

from .set_on import SetOnResponse
from .set_off import SetOffResponse
from .pulse import PulseResponse
from .flashing import FlashingResponse

__all__ = [
    'SetOnResponse',
    'SetOffResponse',
    'PulseResponse',
    'FlashingResponse',
]
